@@ -1,14 +1,15 @@
-from functools import wraps
-import threading
-import time
 import datetime
 import logging
-from typing import Optional, Iterable, Dict, List
+import threading
+import time
 from collections import OrderedDict
+from functools import wraps
+from typing import Dict, Iterable, List, Optional
 
 import binsync.data
-from ..client import Client
-from ..data import User, Function, StackVariable, Comment, Struct, GlobalVariable, Enum
+from binsync.core.client import Client, SchedSpeed
+from binsync.data import (Comment, Enum, Function, GlobalVariable,
+                          StackVariable, Struct, User)
 
 _l = logging.getLogger(name=__name__)
 
@@ -43,7 +44,7 @@ def make_state(f):
 
         kwargs['state'] = state
         r = f(self, *args, **kwargs)
-        self.client.commit_state(msg=self._generate_commit_message(f, *args, **kwargs))
+        self.client.commit_state(state, msg=self._generate_commit_message(f, *args, **kwargs))
         return r
 
     return state_check
@@ -70,7 +71,7 @@ def make_state_with_func(f):
 
         kwargs['state'] = state
         r = f(self, *args, **kwargs)
-        self.client.commit_state(msg=self._generate_commit_message(f, *args, **kwargs))
+        self.client.commit_state(state, msg=self._generate_commit_message(f, *args, **kwargs))
         return r
 
     return _make_state_with_func
@@ -99,7 +100,8 @@ def make_ro_state(f):
 # Description Constants
 #
 
-BINSYNC_RELOAD_TIME = 10
+# https://stackoverflow.com/questions/10926328
+BUSY_LOOP_COOLDOWN = 0.5
 
 
 class SyncControlStatus:
@@ -128,8 +130,9 @@ class BinSyncController:
     The client will be set on connection. The ctx_change_callback will be set by an outside UI
 
     """
-    def __init__(self, headless=False):
+    def __init__(self, headless=False, reload_time=10):
         self.headless = headless
+        self.reload_time = reload_time
 
         # client created on connection
         self.client = None  # type: Optional[Client]
@@ -158,13 +161,7 @@ class BinSyncController:
         with self.queue_lock:
             self.cmd_queue[time.time()] = (cmd_func, args, kwargs)
 
-    def _eval_cmd_queue(self):
-        cmd = None
-        with self.queue_lock:
-            if len(self.cmd_queue) > 0:
-                # pop the first command from the queue
-                cmd = self.cmd_queue.popitem(last=False)[1]
-
+    def _eval_cmd(self, cmd):
         # parse the command if present
         if not cmd:
             return
@@ -172,32 +169,47 @@ class BinSyncController:
         func, f_args, f_kargs = cmd[:]
         func(*f_args, **f_kargs)
 
+    def _eval_cmd_queue(self):
+        with self.queue_lock:
+            if not self.cmd_queue:
+                return
+
+            job_count = 1
+            jobs = [
+                self.cmd_queue.popitem(last=False)[1] for _ in range(job_count)
+            ]
+
+        for job in jobs:
+            self._eval_cmd(job)
+
     def updater_routine(self):
         while True:
-            time.sleep(1)
+            time.sleep(BUSY_LOOP_COOLDOWN)
 
-            # verify the client is connected
+            # validate a client is connected to this controller (may not have remote )
             if not self.check_client():
                 continue
 
+            # do git pull/push operations if a remote exist for the client
+            if self.client.has_remote:
+                if self.client.last_pull_attempt_ts is None:
+                    self.client.update(commit_msg="User created")
+
+                # update every reload_time
+                elif time.time() - self.client.last_pull_attempt_ts > self.reload_time:
+                    self.client.update()
+
             if not self.headless:
-                # update context knowledge every 1 second
+                # update context knowledge every loop iteration
                 if self.ctx_change_callback:
                     self._check_and_notify_ctx()
 
                 # update the control panel with new info every BINSYNC_RELOAD_TIME seconds
                 if self._last_reload is None or \
-                        (datetime.datetime.now() - self._last_reload).seconds > BINSYNC_RELOAD_TIME:
-                    self._last_reload = datetime.datetime.now()
+                        time.time() - self._last_reload > self.reload_time:
+                    self._last_reload = time.time()
                     self._update_ui()
 
-            # update client every BINSYNC_RELOAD_TIME seconds if it has a remote connection
-            if self.client.has_remote and (
-                    (self.client.last_pull_attempt_at is None) or
-                    (datetime.datetime.now() - self.client.last_pull_attempt_at).seconds > BINSYNC_RELOAD_TIME
-            ):
-                self.client.update()
-
             # evaluate commands started by the user
             self._eval_cmd_queue()
 
@@ -229,6 +241,7 @@ class BinSyncController:
             user, path, binary_hash, init_repo=init_repo, remote_url=remote_url
         )
 
+        _l.info("Starting the updater thread in the controller")
         self.start_updater_routine()
         return self.client.connection_warnings
 
@@ -255,8 +268,12 @@ class BinSyncController:
         self.headless = not self.headless
 
     @init_checker
-    def users(self) -> Iterable[User]:
-        return self.client.users()
+    def users(self, priority=None) -> Iterable[User]:
+        return self.client.users(priority=priority)
+
+    def usernames(self, priority=None) -> Iterable[str]:
+        for user in self.users(priority=priority):
+            yield user.name
 
     #
     # Override Mandatory Functions
@@ -297,6 +314,15 @@ class BinSyncController:
         """
         raise NotImplementedError
 
+    def goto_address(self, func_addr) -> None:
+        """
+        Relocates decompiler display to provided address
+
+        @param func_addr:
+        @return:
+        """
+        raise NotImplementedError
+
     #
     # Fillers
     #
@@ -339,6 +365,14 @@ class BinSyncController:
         """
         raise NotImplementedError
 
+    @init_checker
+    @make_ro_state
+    def fill_global_vars(self, user=None, state=None):
+        for off, gvar in state.global_vars.items():
+            self.fill_global_var(off, user=user, state=state)
+
+        return True
+
     @init_checker
     @make_ro_state
     def fill_enum(self, enum_name, user=None, state=None):
@@ -370,6 +404,150 @@ class BinSyncController:
         """
         raise NotImplementedError
 
+    def fill_functions(self, user=None, state=None):
+        change = False
+        for addr, func in state.functions.items():
+            change |= self.fill_function(addr, user=user, state=state)
+
+        return change
+
+    @init_checker
+    @make_ro_state
+    def fill_all(self, user=None, state=None):
+        """
+        Connected to the Sync All action:
+        syncs in all the data from the targeted user
+
+        TODO:
+        - add support for enums
+
+        @param user:
+        @param state:
+        @param no_functions:
+        @return:
+        """
+        _l.info(f"Filling all data from user {user}...")
+
+        fillers = [
+            self.fill_structs, self.fill_enums, self.fill_global_vars
+        ]
+
+        for filler in fillers:
+            filler(user=user, state=state)
+
+    @init_checker
+    def magic_fill(self, preference_user=None):
+        """
+        Traverses all the data in the BinSync repo, starting with an optional preference user,
+        and sequentially merges that data together in a non-conflicting way. This also means that the prefrence
+        user makes up the majority of the initial data you sync in.
+
+        This process supports: functions (header, stack vars), structs, and global vars
+        TODO:
+        - support for comments
+        - support for enums
+        - refactor fill_function to stop attempting to set master state after we do
+        -
+
+        @param preference_user:
+        @return:
+        """
+
+        _l.info(f"Staring a magic sync with a preference for {preference_user}")
+        # re-order users for the prefered user to be at the front of the queue (if they exist)
+        all_users = list(self.usernames(priority=SchedSpeed.FAST))
+        preference_user = preference_user if preference_user else self.client.master_user
+        all_users.remove(preference_user)
+        master_state = self.client.get_state(user=self.client.master_user, priority=SchedSpeed.FAST)
+
+        #
+        # structs
+        #
+
+        _l.info(f"Magic Syncing Structs...")
+        pref_state = self.client.get_state(user=preference_user, priority=SchedSpeed.FAST)
+        for struct_name in self.get_all_changed_structs():
+            _l.info(f"Looking at strunct {struct_name}")
+            pref_struct = pref_state.get_struct(struct_name)
+            for user in all_users:
+                user_state = self.client.get_state(user=user, priority=SchedSpeed.FAST)
+                user_struct = user_state.get_struct(user)
+
+                if not user_struct:
+                    continue
+
+                if not pref_struct:
+                    pref_struct = user_struct.copy()
+                    continue
+
+                pref_struct = Struct.from_nonconflicting_merge(pref_struct, user_struct)
+                pref_struct.last_change = None
+
+            if pref_struct:
+                master_state.structs[struct_name] = pref_struct
+
+            self.fill_struct(struct_name, state=master_state)
+        self.client.commit_state(master_state, msg="Magic Sync Structs Merged")
+
+        #
+        # functions
+        #
+
+        master_state = self.client.get_state(user=self.client.master_user, priority=SchedSpeed.FAST)
+
+        _l.info(f"Magic Syncing Functions...")
+        pref_state = self.client.get_state(user=preference_user, priority=SchedSpeed.FAST)
+        for func_addr in self.get_all_changed_funcs():
+            _l.info(f"Looking at func {hex(func_addr)}")
+            pref_func = pref_state.get_function(addr=func_addr)
+            for user in all_users:
+                user_state = self.client.get_state(user=user, priority=SchedSpeed.FAST)
+                user_func = user_state.get_function(func_addr)
+
+                if not user_func:
+                    continue
+
+                if not pref_func:
+                    pref_func = user_func.copy()
+                    continue
+
+                pref_func = Function.from_nonconflicting_merge(pref_func, user_func)
+                pref_func.last_change = None
+
+            master_state.functions[func_addr] = pref_func
+            self.fill_function(func_addr, state=master_state)
+
+        self.client.commit_state(master_state, msg="Magic Sync Funcs Merged")
+
+        #
+        # global vars
+        #
+
+        _l.info(f"Magic Syncing Global Vars...")
+        master_state = self.client.get_state(user=self.client.master_user, priority=SchedSpeed.FAST)
+        pref_state = self.client.get_state(user=preference_user, priority=SchedSpeed.FAST)
+        for gvar_addr in self.get_all_changed_global_vars():
+            pref_gvar = pref_state.get_global_var(gvar_addr)
+            for user in all_users:
+                user_state = self.client.get_state(user=user, priority=SchedSpeed.FAST)
+                user_gvar = user_state.get_global_var(gvar_addr)
+
+                if not user_gvar:
+                    continue
+
+                if not pref_gvar:
+                    pref_gvar = user_gvar.copy()
+                    continue
+
+                pref_gvar = GlobalVariable.from_nonconflicting_merge(pref_gvar, user_gvar)
+                pref_gvar.last_change = None
+
+            master_state.global_vars[gvar_addr] = pref_gvar
+            self.fill_global_var(gvar_addr, state=master_state)
+
+        self.client.commit_state(master_state, msg="Magic Sync Global Vars Merged")
+        _l.info(f"Magic Syncing Completed!")
+
     #
     # Pushers
     #
@@ -410,7 +588,7 @@ class BinSyncController:
 
     @init_checker
     @make_ro_state
-    def pull_function(self, func_addr, user=None, state=None) -> Function:
+    def pull_function(self, func_addr, user=None, state=None) -> Optional[Function]:
         if not func_addr:
             return None
 
@@ -541,3 +719,30 @@ class BinSyncController:
         if not sync_data:
             msg = "Generic Update"
         return msg
+
+    def get_all_changed_funcs(self):
+        known_funcs = set()
+        for username in self.usernames(priority=SchedSpeed.FAST):
+            state = self.client.get_state(user=username, priority=SchedSpeed.FAST)
+            for func_addr in state.functions:
+                known_funcs.add(func_addr)
+
+        return known_funcs
+
+    def get_all_changed_structs(self):
+        known_structs = set()
+        for username in self.usernames(priority=SchedSpeed.FAST):
+            state = self.client.get_state(user=username, priority=SchedSpeed.FAST)
+            for struct_name in state.structs:
+                known_structs.add(struct_name)
+
+        return known_structs
+
+    def get_all_changed_global_vars(self):
+        known_gvars = set()
+        for username in self.usernames(priority=SchedSpeed.FAST):
+            state = self.client.get_state(user=username, priority=SchedSpeed.FAST)
+            for offset in state.global_vars:
+                known_gvars.add(offset)
+
+        return known_gvars
