@@ -1,8 +1,10 @@
-from .artifact import Artifact
-from typing import Dict
 from collections import OrderedDict
+from typing import Dict
+
 import toml
 
+from binsync.data.artifact import Artifact
+
 
 class Enum(Artifact):
     __slots__ = Artifact.__slots__ + (
@@ -40,3 +42,10 @@ class Enum(Artifact):
         for name, enum in enums.items():
             enums_[name] = enum.__getstate__()
         return enums_
+
+    def copy(self):
+        return Enum(
+            self.name,
+            self.value_map.copy(),
+            last_change=self.last_change
+        )
