@@ -1,6 +1,7 @@
-from .artifact import Artifact
 import toml
 
+from binsync.data.artifact import Artifact
+
 
 class GlobalVariable(Artifact):
     __slots__ = Artifact.__slots__ + (
@@ -41,3 +42,7 @@ class GlobalVariable(Artifact):
         for v in sorted(global_vars.values(), key=lambda x: x.addr):
             global_vars_["%x" % v.addr] = v.__getstate__()
         return global_vars_
+
+    def copy(self):
+        gvar = GlobalVariable(self.addr, self.name, self.type_str, self.size)
+        return gvar
